# rrxutil.rb: (coding: Windows-31J) 2013/11/09
  # add some methods for utilizing Kramdown to Rrx module
  # Copyright (C) T. Yoshiizumi, 2013 All rights reserved.
require "tempfile"
require "nkf"
require "rrxwin"

module Rrx
  module_function()

  def md_table(mm, *arg)  # R's array | matrix to table of markdown
    sep = opt = fsep = type = nil
    if arg.size > 0
      while w = arg.shift
        if (w.class == String and w =~ /^=+$/) or
            (w.class == Array and w.size == 1 and w[0] =~ /^=+$/)
          fsep = ["====="]
          next
        end
        sep = w  if w.class == Array
        opt = w  if w.class == String
        type = w  if w.class == Symbol
      end
    end
    mm = Rrx::to_rba(mm, type)
    if mm[0].class != Array  # not 2d array
      mm2 = [mm]
    elsif mm.size == 1  # 2d array which consists of only one line
      mm2 = mm
    else  # if mm.size > 1
      sep = ["----"] * mm[0].size  unless sep
      mm2 = (sep == []) ? mm : [mm[0], sep] + mm[1..-1]
      if fsep != nil and mm2[-2] != sep
        mm2 = mm2[0..-2] + [fsep, mm2[-1]]
      end
    end
    res = ""
    mm2.each {|row|
      w = row.join("\x7f\x7f")
      w = w.gsub(/\|/, "\\|")
      w = w.gsub(/\x7f\x7f/, "|")
      res += "|#{w}\n"
    }
    opt = opt.to_s.sub(/\s+$/, "")
    if opt != ''
      res += "{: #{opt}}\n"
    end
    return res
  end

  def array2table(aa, *arg)  # R's array to table of markdown
    return md_table(aa, :array, *arg)
  end

  def matrix2table(aa, *arg)  # R's matrix to table of markdown
    return md_table(aa, :matrix, *arg)
  end

  def html_template(enc=nil, ins=nil)  # create a html template
    $rrx_enc = (enc.class == String) ? enc : nil
    if ins == :mathjax
      ins = <<-MATHJAX
		<script type="text/javascript"
		  src="http://cdn.mathjax.org/mathjax/latest/\
		MathJax.js?config=TeX-AMS_HTML">
		</script>
		<meta http-equiv="X-UA-Compatible" content="IE=EmulateIE7">
      MATHJAX
    end
    template_str = <<-EOS
		<% if $rrx_enc != nil
		     if RUBY_VERSION >= '1.9'
		       enc = @converter.root.options[:encoding]
		       enc = Encoding.default_external unless enc
		     end
		     cset = $rrx_enc
		   elsif RUBY_VERSION >= '1.9'
		     enc = @converter.root.options[:encoding]
		     enc = Encoding.default_external unless enc
		     cset = enc.to_s =~ /windows-31j|cp932|sjis|shift.jis/i ?
		         'Shift_JIS' : enc.to_s
		   else
		     enc = $KCODE.to_s[0,1].downcase
		     enc = 'w' unless ['e', 's'].include?(enc)
		     cset = {'e'=>'euc-jp', 's'=>'Shift_JIS', 'w'=>'utf-8'}[enc]
		   end
		%>\
		<!DOCTYPE html>
		<html>
		<head>
		<meta http-equiv="Content-type" content="text/html; charset=<%= cset %>" />
		<% extend ::Kramdown::Utils::Html
		title = ''
		h = @converter.root.children.find {|c| c.type == :header}
		if h
		  collector = lambda {|c| c.children.collect {|cc| cc.type == :text ?
		  escape_html(cc.value, :text) : collector.call(cc)}.join('')}
		  title = collector.call(h)
		  title = title.to_s.encode(enc) if RUBY_VERSION >= '1.9'
		end %>\
		<title><%= title %></title>
		<meta name="generator" content="kramdown <%= ::Kramdown::VERSION %>" />
		#{ins}</head>
		<body>
		<%= @body %>
		</body>
		</html>
    EOS
    template_str = template_str.gsub(/\t\t/, "")
    ff = Tempfile.new("tpl_")
    ff.write template_str
    ff.close
    return ff
  end

  def tex_template(dclass=nil)  # create a LaTeX template file(temporary)
    dclass = nil  if dclass.class != String
    dclass = !dclass ? '\documentclass{jarticle}' : dclass.chomp
    template_str = <<-EOS
		#{dclass}
		\\usepackage{listings}
		\\usepackage{hyperref}
		<% @converter.data[:packages].each {|pkg| %>\
		\\usepackage{<%= pkg %>}
		<% } %>\
		<% if @converter.data[:packages].include?('fancyvrb') %>\
		\\VerbatimFootnotes
		<% end %>\
		<% if @converter.data[:packages].include?('acronym') %>\
		<%   @converter.root.options[:abbrev_defs].each_pair do |k,v| %>\
		\\acrodef{<%= @converter.normalize_abbreviation_key(k) %>}[\
		<%= k %>]{<%= @converter.escape(v) %>}
		<%   end %>\
		<% end %>\
		\\hypersetup{colorlinks=true,urlcolor=blue}
		\\begin{document}
		<%= @body %>
		\\end{document}
    EOS
    template_str = template_str.gsub(/\t\t/, "")
        ff = Tempfile.new("tpl_")
    ff.write template_str
    ff.close
    return ff
  end
end
