#! ruby -Ku
  # templateでstand-aloneのLaTeXを作成  (coding: utf-8)
require "rubygems"
require "kramdown"
require "spreadsheet"
require "rrxutil"
        ## Rプログラムの設定
rpro = <<EOS
xx <- rbind(c("東京都", 13216221, 2188.67, 6038.47),
    c("神奈川県", 9072133, 2415.86, 3755.24),
    c("大阪府", 8863324, 1901.42, 4661.42),
    c("愛知県", 7425952, 5165.14, 1437.71),
    c("全国", 127532028, 372923.77, 341.98))
colnames(xx) <- c("都道府県", "人口", "面積", "人口密度")
robj(xx)
EOS
        ## Rの実行，LaTeXへの変換
hs = Rrx.rexec(rpro)  # Rの実行
xx = hs["robj1"]["self"]
md_str = <<EOS2
# 都道府県の人口・面積・人口密度\n
　都道府県の人口(人)、面積(平方km)、人口密度(人/平方km)は次のとおり。\n
　人口の多い順に、４都府県のみ示す。\n
#{Rrx.matrix2table(xx, ["==="], 'border="1"')}
# ぜんぜん関係ないけど数式の例
分数の計算：
$$ \\frac{1}{2} + \\frac{1}{3} = \\frac{3 + 2}{6} = \\frac{5}{6} $$\n
相加平均の計算は次のとおり。\n
$$
\\begin{equation*}
μ = \\frac{1}{n}\\sum_{i=1}^{n} x_i = 
\\frac{x_1 + x_2 + \\cdots + x_n}{n}
\\end{equation*}
$$
EOS2
        ## kramdownによるLaTeXの生成
tfile = Rrx.tex_template(<<'EOS3')
\documentclass{jsarticle}
\renewcommand{\presectionname}{第}
\renewcommand{\postsectionname}{章}
EOS3
kdoc = Kramdown::Document.new(md_str, :template=>tfile.path)
tex_str = kdoc.to_latex
File.open("result.tex", "w") {|ff|  ff.write tex_str}
tfile.unlink  # テンプレートファイルを削除
puts "'result.tex' を出力しました."
