#! ruby -Ku
  # create a html document by R and kramdown (coding: utf-8)
require "rubygems"
require "roo"
require "kramdown"
require "erb"
require "rrxutil"
        ## Rプログラムの記述
rpro = <<R_PROGRAM
source("http://cup.sakura.ne.jp/R/tjsf.r", encoding="utf-8")
dtf <- read.csv("@@last@@", header=T)
ma_col <- grep("に期待$", colnames(dtf))  # 複数回答の列番号群を取得
dtf <- ma.refactor(dtf, ma_col, c("はい", "ノーマーク"))
ww <- sa1(dtf, "年齢")
png(file="bar.png")
par(family = "Japan1GothicBBB")  # 日本語フォントにゴチックを指定
barplot(ww$tbl1, main="年齢別回答者数", xlab="年齢階層",
    ylab="回答者数", names.arg=names(ww$tbl1))
dev.off()
ss <- join.tbl(ww$tbl2, ww$pct1, "%d(%.1f)")  # 年齢の構成分布
ww <- ma1(dtf, "年齢", ma_col)
tt <- join.tbl(ww$tbl, ww$pct, "%d(%.1f)")  # 期待事項の回答状況
xx <- ww$pct
xx <- xx[,-ncol(xx)]  # 右端の「該当数」を削除
nr <- nrow(xx)  # 最下行の「合計」の行番号
for (j in 1:ncol(xx))  # 特化係数の算出
  for (i in 1:(nr-1)) {
    if (xx[nr,j] == 0)  next
    xx[i,j] <- xx[i,j] / xx[nr,j]
  }
robj(list(回答者分布=ss, 回答状況=tt, 特化係数=round(xx,2)))
R_PROGRAM
        ## Rの実行と結果の受け取り
xl_file = "http://www.timecube.co.jp/Kaleid/aplcn/apl1/sample1.xls"
sheets = Rrx.xls2ary(xl_file)  # Excelファイルを読み込んで配列に変換
Rrx.temp_make(sheets[6], ".csv")  # 第７ワークシートをcsvとして出力
hs = Rrx.rexec(rpro)  # Rの実行
xx = hs["robj1"]
        ## markdown原稿の生成
tblopt = 'border="1"'  # <table>のオプション(枠線の表示)
md_src = <<MARKDOWN
# 新車発表会のアンケート結果について\n
　○○年××日に行われた新車発表会で、来場者に対して行った
アンケート(回答者500人)の考察結果を記す。\n
- - - - - - - - -\n
# ＜目次＞
{:#contents}\n
* contents list
{:toc}\n
- - - - - - - - -\n
## １．　「年齢」と「何に期待して来場したか」との関連\n
### 1-1　回答者の年齢の分布\n
　まず、回答者の年齢の分布をみる。回答者総数は500人。\n
<%= Rrx.array2table(xx["回答者分布"], tblopt) %>
![画像：年齢別回答者数](bar.png)\n
　概観を箇条書きにすると次のとおり。\n
- 20歳未満と60歳以上が両者とも13人(2.6%)と少数。
- 40代が163人と多く、約３分の１を占める。
- 20代・30代・50代は、いずれも２割前後を占める。\n
[△ 目次に戻る](#contents)\n
- - - - - - - - -\n
### 1-2　回答状況\n
　次に、「外観に期待」などの期待事項に「はい」と回答した人の数と
その比率をみると下のとおり。\n
　全体的にみると、「居住性に期待」と「価格に期待」が53%程度を占め、
「外観に期待」(36%)と「乗心地に期待」(41%)に比べて高い。\n
<%= Rrx.matrix2table(xx["回答状況"], tblopt, "===") %>
[△ 目次に戻る](#contents)\n
- - - - - - - - -\n
### 1-3　特化係数\n
　先の表だと、年齢層による特徴を把握しにくいので、「合計」欄の
比率を基準とする特化係数をみてみよう。\n
　なお、「合計」欄の数値は特化係数ではなく、回答比率(％)である。\n
<%= Rrx.matrix2table(xx["特化係数"], tblopt, "===") %>
　特化係数を手がかりにして大雑把に概観すると、次の点を指摘できる。\n
- 全体的に高めの「居住性」と「価格」は、どの年齢層別も比較的高いか平均的。
- 「外観」は、30代と50代が高く、40代が平均的、60歳以上と29歳以下は低い。
- 「乗心地」は、40代と60歳以上が高く、19歳以下と30代が低い。
20代と50代は平均的。\n
　「何に期待して来場したか」について、年齢階層別の特徴を上げたが、
「年齢が高いほど……」といった一定方向の傾向は指摘できない。\n
　ちなみに、各々の期待事項ごとに年齢層とのクロス集計を行い、
\\\\(χ^{2}\\\\)検定にかけた結果、いずれも有意性は認められなかった。\n
[△ 目次に戻る](#contents)\n
- - - - - - - - -\n
## ２．　「運転歴」と「何に期待して来場したか」との関連\n
　［以下、省略］\n
MARKDOWN
        ## markdown→html変換
md_str = ERB.new(md_src).result(binding)
File.open("result.md", "w") {|ff|  ff.write md_str}
enc = "euc-jp"  # これから作るhtmlのencode
tfile = Rrx.html_template(enc, :mathjax)  # テンプレート生成(tempfile)
opt = {:template=>tfile.path, :toc_levels=>[2,3]}
kdoc = Kramdown::Document.new(md_str, opt)
html_str = Rrx.nkfex(kdoc.to_html, enc)
File.open("result.htm", "w") {|ff|  ff.write html_str}
tfile.unlink  # テンプレートファイルを削除
puts "'result.md', 'result.htm', 'bar.png' を出力しました."
